#!/bin/bash

# Set up experiment data
source setup.sh

# Generate data
folder="../results/alloverheads"

# Constant:
# * CS alpha (fraction not at L_max)
# * CS beta (fraction of L_max for shorter requests)
# * # resources
# * # resources per request
# * critical section length
cs_alpha=0
cs_beta=100

# Varying:
# * # tasks
# * algorithm

# Measure:
# * overhead
# * blocking

echo -n "Running experiments for ECRTS 2019 - all overheads"

for cslength in 40
do
    for numresourcesnested in 4
    do
        for numresources in 64
        do
            for alg in 1 2 3 8 13 14 15 16 17 18 19 20 21
            do
                # One core per task
                taskcounts=(${taskcounts[@]})
                exe="./../main"

                for (( i = 0; i < ${#taskcounts[@]}; i++ ))
                do
                    numtasks=${taskcounts[$i]}

                    if [[ $alg -eq 13 ]]
                    then
                        numres=3
                        reqvar=1
                    elif [[ $alg -eq 14 ]]
                    then
                        numres=4
                        reqvar=1
                    elif [[ $alg -eq 15 ]]
                    then
                        numres=5
                        reqvar=1
                    elif [[ $alg -eq 16 ]]
                    then
                        numres=6
                        reqvar=1
                    elif [[ $alg -eq 17 ]]
                    then
                        numres=7
                        reqvar=1
                    elif [[ $alg -eq 18 ]]
                    then
                        numres=8
                        reqvar=1
                    elif [[ $alg -eq 19 ]]
                    then
                        numres=9
                        reqvar=1
                    elif [[ $alg -eq 20 ]]
                    then
                        numres=10
                        reqvar=1
                    elif [[ $alg -eq 21 ]]
                    then
                        numres=2
                        reqvar=1
                    else
                        numres=$numresources
                        reqvar=$numresourcesnested
                    fi

                    if [[ $verbose -eq 1 ]]
                    then
                        currenttime="`date +%H:%M:%S`"
                        echo "[$currenttime] n=$numtasks ($numcores cores), q=$numres, cslength=$cslength, k=$reqvar, alg=$alg"
                    else
                        echo -n "."
                    fi

                    $exe $numtasks $reqvar $numres $cslength $cs_alpha $cs_beta $alg $folder/$alg/$numtasks-$numresources-$numresourcesnested-$cslength-$alg.csv
                done
            done
        done
    done
done

if [[ $verbose -eq 0 ]]
then
    echo ""
fi
