#include "sharedres.h"
#include "blocking.h"
#include "rw-blocking.h"
#include <iostream>

#include "stl-helper.h"

bool share_resources2(const std::vector<RequestBound> r_list1, const std::vector<RequestBound> r_list2)
{
    unsigned int i, j;
    for(i = 0; i < r_list1.size(); i++)
    {
        for(j = 0; j < r_list2.size(); j++)
        {
            if(r_list1[i].get_resource_id() == r_list2[j].get_resource_id())
            {
                return true;
            }
        }
    }
    return false;
}

unsigned int get_contention2(const TaskInfo &tsk,
                             const std::vector<TaskInfo> &all_tasks)
{
    std::vector<const RequestBound*> conflicting_reqs;
    unsigned int i;
    unsigned int c = 0;
    for(i = 0; i < all_tasks.size(); i++)
    {
        if(tsk.get_id() != all_tasks[i].get_id())
        {
            if(share_resources2(tsk.get_requests(), all_tasks[i].get_requests()))
            {
                c++;
            }
        }
    }
    return c;
}

BlockingBounds* gcrnlp_bounds(const ResourceSharingInfo& info,
				 unsigned int procs_per_cluster,
				 int dedicated_irq)
{
    // split everything by partition
    Clusters clusters;
    split_by_cluster(info, clusters);

    // split each partition by resource
    ClusterResources resources;

    split_by_resource(clusters, resources);

    // sort each contention set by request length
    sort_by_request_length(resources);

    std::vector<const RequestBound*> all_reqs;
    unsigned int k = 0;
    for(k = 0; k < resources[0].size(); k++)
    {
        foreach(resources[0][k], it)
        {
            const RequestBound* req = *it;
            unsigned int m = 0;
            bool req_already_recorded = false;
            for(m=0; m<all_reqs.size(); m++)
            {
                if(all_reqs[m]->get_task() == req->get_task())
                {
                    req_already_recorded = true;
                    break;
                }
            }
            if (req_already_recorded == false)
            {
                all_reqs.push_back(req);
            }
        }
    }
    sort_by_request_length(all_reqs);


    // We need for each task the maximum request span.  We also need the
    // maximum direct blocking from remote partitions for each request. We
    // can determine both in one pass.

    unsigned int i;

    // direct blocking results
    BlockingBounds* _results = new BlockingBounds(info);
    BlockingBounds& results = *_results;

    for (i = 0; i < info.get_tasks().size(); i++)
    {
        const TaskInfo& tsk  = info.get_tasks()[i];
        Interference bterm;

        if (tsk.get_requests().size() > 0)
        {
            const RequestBound* this_req = &tsk.get_requests()[0];

            Interference wblocking, wblocking_w1;

            int idx;
            unsigned int parallelism = procs_per_cluster;
	    enumerate(resources, ct, idx)
            {
                parallelism = procs_per_cluster;

                if (idx == dedicated_irq)
                {
                    parallelism--;
                }

                if (parallelism && (int) tsk.get_cluster() == idx)
                {
                    parallelism--;
                }
            }

            // Calculate blocking
            if (parallelism == 0)
            {
                wblocking = 0;
            }
            else
            {
                unsigned int contention = get_contention2(tsk,info.get_tasks());
                unsigned int count = 0;
                unsigned int total_length = 0;
                // Goal (Ci)(Lmax) + (Ci)(Li)
                unsigned int num_contending_remaining = contention;

                // Keep track of when the requests we're considering are
                // now shorter than the current task's request
                bool seen_current_task = false;

                foreach(all_reqs, it)
                {
                    const RequestBound* req = *it;
                    if(num_contending_remaining == 0 || parallelism - count == 0)
                    {
                        // If we are in the case where we have an abundance of other
                        // cores, and we're out of num_contending_remaining, we're done!
                        if (parallelism >= 2 * contention)
                        {
                            break;
                        }

                        // Otherwise, if we have run out of available cores, we are
                        // also done
                        if (parallelism - count == 0)
                        {
                            break;
                        }

                        // If we're still here, it's the case where we are somewhat
                        // (or very) limited on cores, and we haven't used all of them
                        // yet, but we have run out of remaining contending requests,
                        // so we should fill them with gaps
                        unsigned int num_gaps = std::min(parallelism - count, count); // it's actually always parallelism-count...
                        total_length += num_gaps*this_req->get_request_length();
                        count += num_gaps;
                        break;
                    }

                    if (parallelism >= 2 * contention)
                    {
                        // If the number of other cores is at least 2*contention, then
                        // we always just add another request + a gap (Li) in a pair
                        if(req->get_task() != &tsk)
                        {
                            // If I don't contend with this request, it does not contribute to the Ci*Lmax term
                            if (!share_resources2(tsk.get_requests(), req->get_task()->get_requests())) {
                                continue;
                            }

                            // The at-most Ci tasks larger than me contribute to the
                            // blocking term - Ci*Lmax
                            unsigned int num;
                            num = 1; // FIFO constraint means at most 1 request per task,
                                     // and num_contending_remaining is >0 if we're here
                            total_length += num*req->get_request_length();
                            count += num;
                            num_contending_remaining -= num;

                            // Also add the gap: Ci*Li
                            total_length += num*this_req->get_request_length();
                            count += num;
                        }

                    }
                    else
                    {
                        // Otherwise, we need to make sure to add as many larger requests
                        // as possible, then add gaps, then add remaining pairs, with Lmax
                        // of the other request smaller than me (all in this order!)
                        if(req->get_task() != &tsk)
                        {
                            // If I don't contend with this request, it does not contribute to the Ci*Lmax term
                            if (!share_resources2(tsk.get_requests(), req->get_task()->get_requests())) {
                                continue;
                            }

                            if (!seen_current_task)
                            {
                                // The at-most Ci tasks larger than me contribute to the
                                // blocking term - Ci*Lmax
                                unsigned int num;
                                num = 1; // FIFO constraint means at most 1 request per task,
                                         // and num_contending_remaining is >0 if we're here
                                total_length += num*req->get_request_length();
                                count += num;
                                num_contending_remaining -= num;
                            }
                            else
                            {
                                // The task smaller than me contributes to the blocking term
                                // Ci*Lmax
                                unsigned int num;
                                num = 1; // FIFO constraint means at most 1 request per task,
                                         // and num_contending_remaining is >0 if we're here
                                total_length += num*req->get_request_length();
                                count += num;
                                num_contending_remaining -= num;

                                // Also add the gap: Ci*Li
                                // If there are still available cores, add a gap of Li
                                if (parallelism - count > 0)
                                {
                                    total_length += num*this_req->get_request_length();
                                    count += num;
                                }
                            }
                        }
                        else
                        {
                            // My task is included in the top Ci requests, so any task
                            // seen so far has a larger CS length, and thus could have
                            // had a gap I can't quite fit in (the Ci * Li term)
                            unsigned int num_gaps = std::min(parallelism - count, count);
                            total_length += num_gaps*this_req->get_request_length();
                            count += num_gaps;

                            // Mark that I've seen myself in the list, so nothing else
                            // will have gaps that I could have fit in (they're all
                            // shorter than me, so nothing else could come in to add
                            // to my delay)
                            seen_current_task = true;
                        }
                    }
                }
                wblocking.total_length = total_length;
                wblocking.count = count;
            }

            wblocking_w1 = wblocking;
            wblocking_w1.total_length += this_req->get_request_length();
            wblocking_w1.count += 1;

            results.raise_request_span(i, wblocking_w1);
            bterm += wblocking;
        }
        results[i] = bterm;
    }

    charge_arrival_blocking(info, results);

    return _results;
}
