#include "sharedres.h"
#include "blocking.h"
#include "rw-blocking.h"
#include <iostream>

#include "stl-helper.h"


BlockingBounds* new_pftl_bounds(const ResourceSharingInfo& info,
					 unsigned int procs_per_cluster,
					 int dedicated_irq)
{
	// split everything by partition
	Clusters clusters;

	split_by_cluster(info, clusters);

	// split each partition by resource
	ClusterResources resources;

	split_by_resource(clusters, resources);

	// split all by resource
	Resources all_task_reqs, all_reads, __all_writes;
	split_by_resource(info, all_task_reqs);
	split_by_type(all_task_reqs, all_reads, __all_writes);

	// sort each contention set by request length
	sort_by_request_length(resources);
	sort_by_request_length(all_reads);

	// split by type --- sorted order is maintained
	ClusterResources __reads, writes;
	split_by_type(resources, __reads, writes);

                                    std::vector<const RequestBound*> write_reqs;
                                    unsigned int k = 0;
                                    for(k=0; k<writes[0].size(); k++)
                                    {
                                        foreach(writes[0][k], it){
                                            const RequestBound* req = *it;
                                            unsigned int m = 0;
                                            bool req_already_recorded = false;
                                            for(m=0; m<write_reqs.size(); m++){
                                                if(write_reqs[m]->get_task() == req->get_task()){
                                                    req_already_recorded = true;
                                                    break;
                                                }
                                            }
                                            if (req_already_recorded == false){
                                                write_reqs.push_back(req);
                                            }
                                        }
                                    }
                                    sort_by_request_length(write_reqs);

//std::cout << "-Writes-\n";
//foreach(write_reqs, it)
//{
//    const RequestBound* req = *it;
//    std::cout <<req->get_request_length() << "\n";
//}
//std::cout << "---\n";
                                    std::vector<const RequestBound*> read_reqs;
                                    for(k=0; k<all_reads.size(); k++)
                                    {
                                        foreach(all_reads[k], it){
                                            const RequestBound* req = *it;
                                            unsigned int m = 0;
                                            bool req_already_recorded = false;
                                            for(m=0; m<read_reqs.size(); m++){
                                                if(read_reqs[m]->get_task() == req->get_task()){
                                                    req_already_recorded = true;
                                                    break;
                                                }
                                            }
                                            if (req_already_recorded == false){
                                                read_reqs.push_back(req);
                                            }
                                        }
                                    }
                                    sort_by_request_length(read_reqs);
//std::cout << "-Reads-\n";
//foreach(read_reqs, it)
//{
//    const RequestBound* req = *it;
//    std::cout <<req->get_request_length() << "\n";
//}
//std::cout << "---\n";

	// We need for each task the maximum request span.  We also need the
	// maximum direct blocking from remote partitions for each request. We
	// can determine both in one pass.

	unsigned int i;
        unsigned int one = 1;

	// direct blocking results
	BlockingBounds* _results = new BlockingBounds(info);
	BlockingBounds& results = *_results;

	for (i = 0; i < info.get_tasks().size(); i++)
	{
		const TaskInfo& tsk  = info.get_tasks()[i];
		Interference bterm;

                if (tsk.get_requests().size() > 0)
		{
                        const RequestBound* this_req = &tsk.get_requests()[0];

			Interference wblocking,  rblocking;
			Interference rblocking_w1, wblocking_w1;
			Interference rblocking_r1, wblocking_r1;

                                int idx;
                                unsigned int parallelism = procs_per_cluster;
	                        enumerate(writes, ct, idx)
		                {
                                    parallelism = procs_per_cluster;

		                    if (idx == dedicated_irq)
			                parallelism--;

		                    if (parallelism && (int) tsk.get_cluster() == idx)
			                parallelism--;
		                }
                        // Handle read request
                        if (this_req->is_read())
                        {
			        // Calculate wblocking
                                if (parallelism == 0)
                                    wblocking = 0;
                                else
                                {
                                    unsigned int remaining = 1;
                                    unsigned int count = 0;
                                    unsigned int total_length = 0;
                                    foreach(write_reqs, it)
                                    {
                                        const RequestBound* req = *it;
                                        if (!remaining)
                                            break;
                                        if (req->get_task() != &tsk)
                                        {
                                            total_length += req->get_request_length();
                                            count += 1;
                                            remaining -= 1;
                                        }
                                    }
                                    wblocking.total_length = total_length;
                                    wblocking.count = count;
                                }

			        // Calculate rblocking
                                unsigned int rlimit = std::min(wblocking.count, one);
                                if (rlimit == 0)
                                {
                                        rblocking = 0;
                                }
                                else
                                {
                                    unsigned int remaining = 1;
                                    unsigned int count = 0;
                                    unsigned int total_length = 0;
                                    foreach(read_reqs, it)
                                    {
                                       const RequestBound* req = *it;
                                       if (!remaining)
                                           break;
                                       if (req->get_task() != &tsk &&
                                           (procs_per_cluster != 1 ||
                                            req->get_task()->get_cluster() != (&tsk)->get_cluster()))
                                       {
                                           total_length = req->get_request_length();
                                           count += 1;
                                           remaining -= 1;
                                       }
                                    }
                                    rblocking.total_length = total_length;
                                    rblocking.count = count;
                                }
				wblocking_r1 = wblocking;
			        wblocking_r1 += rblocking;
				wblocking_r1.count        += 1;
			        wblocking_r1.total_length += this_req->get_request_length();

			        results.raise_request_span(i, wblocking_r1);

			        wblocking    += rblocking;
			        bterm += wblocking;
                        }
                        // Handle write request
                        else if (!this_req->is_read())
                        {
			    // Calculate wblocking
                            if (parallelism == 0)
                            {
                                wblocking = 0;
                            }
                            else
                            {
                                    unsigned int max_requests_per_source = 1;
                                    unsigned int count = 0;
                                    unsigned int total_length = 0;
                                        unsigned int remaining = parallelism;
                                        foreach(write_reqs, it)
                                        {
                                           const RequestBound* req = *it;
                                           if (!remaining)
                                               break;
                                           if (req->get_task() != &tsk)
                                           {
                                               unsigned int num;
                                               num = std::min(req->get_max_num_requests(tsk.get_response()),
                                                      max_requests_per_source);
                                               num = std::min(num, remaining);
                                               total_length += num*req->get_request_length();
                                               count += num;
                                               remaining -= num;
                                           }
                                        }
                                        wblocking.total_length = total_length;
                                        wblocking.count = count;

                            }
			        // Calculate rblocking
                                unsigned int rlimit = std::min(wblocking.count + 1, parallelism);
                                if (rlimit == 0)
                                {
                                        rblocking = 0;
                                }
                                else
                                {
                                    unsigned int max_requests_per_source =rlimit;
                                    unsigned int count = 0;
                                    unsigned int total_length = 0;
                                    unsigned int remaining = rlimit;

                                        foreach(read_reqs, it)
                                        {
                                           const RequestBound* req = *it;
                                           if (!remaining)
                                               break;
                                           if (req->get_task() != &tsk &&
                                               (procs_per_cluster != 1 ||
                                                req->get_task()->get_cluster() != (&tsk)->get_cluster()))
                                           {
                                               unsigned int num;
                                               num = std::min(req->get_max_num_requests(tsk.get_response()),
                                                      max_requests_per_source);
                                               num = std::min(num, remaining);
                                               total_length += num*req->get_request_length();
                                               count += num;
                                               remaining -= num;
                                           }
                                        }
                                     rblocking.total_length = total_length;
                                     rblocking.count = count;
                                }

				wblocking_w1 = wblocking;
				rblocking_w1 = rblocking;
				wblocking_w1.total_length += this_req->get_request_length();
				wblocking_w1.count        += 1;

			        // combine
			        wblocking_w1 += rblocking_w1;
			        wblocking_r1 += rblocking_r1;
			        wblocking    += rblocking;

			        results.raise_request_span(i, wblocking_w1);
			        results.raise_request_span(i, wblocking_r1);
			        bterm += wblocking;
                        }
		} // foreach
		results[i] = bterm;
	}

	// This is the initial delay due to priority donation.
	charge_arrival_blocking(info, results);

	return _results;
}

